classdef disturbances < handle
   % disturbances This class contains all the relevant properties of the
   % various disturbances that affect the energy hub components.
   % ------------------------------------------------------------------------
   % This file is part of the EHCM Toolbox v1
   %
   % The EHCM Toolbox - Energy Hub Component Modeling for Model Predictive Control.
   % Copyright (C) 2015  Automatic Control Laboratory, ETH Zurich.
   % 
   % The EHCM Toolbox is free software; you can redistribute it and/or modify
   % it under the terms of the GNU General Public License as published by
   % the Free Software Foundation, either version 3 of the License, or
   % (at your option) any later version.
   % 
   % The EHCM Toolbox is distributed in the hope that it will be useful,
   % but WITHOUT ANY WARRANTY; without even the implied warranty of
   % MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   % GNU General Public License for more details.
   % 
   % You should have received a copy of the GNU General Public License
   % along with the EHCM Toolbox.  If not, see <http://www.gnu.org/licenses/>.
   %
   % ------------------------------------------------------------------------
        
    properties (SetAccess = private)
        rootName
        folderName
        vreal          % real disturbance
        vpred    % predicted disturbance 
        vpredReal
        verror
        verrorReal
        identifiers
    end
    
    methods
        function obj = disturbances(folderName)
            obj.rootName = pwd;
            obj.folderName = [pwd,filesep,folderName];
            obj.vreal = [];
            obj.vpred = [];
            obj.identifiers = [];
            obj.getDisturbances;
            obj.getError;
        end
        
        function getDisturbances(self)
            % Generate disturbance
            
            % Internal Gains
            igreal  = self.getFileClmns({'OccupF', 'EquipF'}, [self.folderName,filesep,'igreal.xlsx']);
            fprintf('IG real extracted\n');
            
            igpredt(:,:,1) = self.getFileClmns({'START'}, [self.folderName,filesep,'igpred_OccupF.xlsx']);
            igpredt(:,:,2) = self.getFileClmns({'START'}, [self.folderName,filesep,'igpred_EquipF.xlsx']);
            
            % get the predictions every 12 hours
            igpred = zeros(size(igpredt,1)/12,72,1);
            cnt = 1;
            for ii = 1:12:size(igpredt,1)
                igpred(cnt,:,:) = igpredt(ii,:,1) + igpredt(ii,:,2);
                cnt = cnt + 1;
            end
            fprintf('IG prediction extracted\n');
            
            
            
            % Weather
            
            wreal=self.getFileClmns({'    TW', '    TA', '   RGN', '   RGE', '   RGS', '   RGW'}, [self.folderName,filesep,'wreal']);
            fprintf('Weather data real extracted\n');
            
            wpred(:,:,1) = self.getFileClmns({'START'}, [self.folderName,filesep,'wpred_TW.xlsx']);
            wpred(:,:,2) = self.getFileClmns({'START'}, [self.folderName,filesep,'wpred_TA.xlsx']);
            wpred(:,:,3) = self.getFileClmns({'START'}, [self.folderName,filesep,'wpred_RGN.xlsx']);
            wpred(:,:,4) = self.getFileClmns({'START'}, [self.folderName,filesep,'wpred_RGE.xlsx']);
            wpred(:,:,5) = self.getFileClmns({'START'}, [self.folderName,filesep,'wpred_RGS.xlsx']);
            wpred(:,:,6) = self.getFileClmns({'START'}, [self.folderName,filesep,'wpred_RGW.xlsx']);
            fprintf('Weather data prediction extracted\n');
            
            % check data integrity for (extremely) false predictions
            for ii = 1:size(wpred,3)
                for jj = 1:size(wpred,2)
                    for kk = 1:size(wpred,1)
                        if ii == 1 || ii == 2 % temperature can not exceed the extremes of sahara and siberia :)
                            if wpred(kk,jj,ii) > 50
                                wpred(kk,jj,ii) = 50;
                            elseif wpred(kk,jj,ii) < -50
                                wpred(kk,jj,ii) = -50;
                            end
                        else     % solar radiation can only be positive
                            if wpred(kk,jj,ii) > 1000
                                wpred(kk,jj,ii) = 1000;
                            elseif wpred(kk,jj,ii) < 0
                                wpred(kk,jj,ii) = 0;
                            end
                        end
                    end
                end
            end
            
            % assign data
            vrealt = zeros(size(wreal,1),7);
            vpredt = zeros(size(wpred,1),72,7);
            identifierst = cell(7,1);
            
            vrealt(:,1) = sum(igreal,2);
            vrealt(:,2:7) = wreal;
            for ii = 1:7
                if ii == 1
                    vpredt(:,:,ii) = igpred(:,:,ii);
                else
                    vpredt(:,:,ii) = wpred(:,:,ii-1);
                end
            end
            vpredtReal = vpredt;
            
            % predictions smoothing (filtering) (median filter to remove
            % spikes and Savitzky-Golay to smooth the predictions)
            
            % As a convenience, you can use the function sgolayfilt to implement
            % a Savitzky-Golay smoothing filter. To use sgolayfilt, you specify 
            % an odd-length segment of the data and a polynomial order strictly 
            % less than the segment length. The sgolayfilt function internally 
            % computes the smoothing polynomial coefficients, performs delay 
            % alignment, and takes care of transient effects at the start and 
            % end of the data record.
            
            for jj = 1:size(vpredt,3)
                for ii = 1:size(vpredt,1)
                    sig1 = vpredt(ii,:,jj);
                    sig1median = medfilt1(sig1,5);  % median filter among 5 samples
                    sig1sgolay   = sgolayfilt(sig1median, 3, 7);    % fit a cubic polynomial to 7 samples
                    if jj == 1 || jj >= 4 % IG and solar radiation can only be positive (may get negative due to polynomial curve fitting)
                        for kk = 1:length(sig1sgolay)
                            if sig1sgolay(kk) < 0
                                sig1sgolay(kk) = 0;
                            end
                        end
                    end
                    vpredt(ii,:,jj) = sig1sgolay;
                end
            end
            
            
            % return values
            
            identifierst{1} = 'v_IG';
            identifierst{2} = {'v_Tgnd','v_AHU_Dwb'};
            identifierst{3} = 'v_Tamb';
            identifierst{4} = 'v_solGlobFac_N';
            identifierst{5} = 'v_solGlobFac_E';
            identifierst{6} = 'v_solGlobFac_S';
            identifierst{7} = 'v_solGlobFac_W';
            
            
            self.vreal = vrealt;
            self.vpredReal = vpredtReal;
            self.vpred = vpredt;
            self.identifiers = identifierst;
        end
        
        function getError(self)
            % function to get the hourly error between predictions and
            % realizations
            vpredt = self.vpred;
            vpredtReal = self.vpredReal;
            
            % create the realization matrix for 12 hour predictions
            mlreal = size(self.vreal,1);
            mlpred = size(vpredt,1);
            vrealt = zeros(size(vpredt));
            for ii = 1:mlpred
                st = (ii-1)*12+1+1; % for the current time we do not have a prediction. (so the realizations start with an additional hour)
                et = (ii-1)*12+72+1;
                if et > mlreal % we do not have so many realization (error assumed zero for these cases)
                    et = mlreal;
                    vrealt(ii,1:mlreal-st+1,:) = self.vreal(st:et,:);
                    vrealt(ii,mlreal-st+2:72,:) = vpredt(ii,mlreal-st+2:72,:);
                else
                    vrealt(ii,:,:) = self.vreal(st:et,:);
                end
            end
            
            errort = vrealt-vpredt;
            errortReal = vrealt-vpredtReal;
            self.verror = errort;
            self.verrorReal = errortReal;
        end
    end
    
    
    methods (Static = true)
        
        function clmns = getFileClmns(startReading, filePath)
            %% read the columns specified in startReading within the filepath xlsx file.
            [~, ~, data_raw]   = xlsread(filePath);
            
            if strcmp(startReading{1}, 'START')
                [rt,ct] = find(strcmp({'START'}, data_raw));
                rt = rt+1;  % data start 2 rows below the 'START' identifier.
                
                clmns2read = 72;    % predictions for 72 hours ahead
                r = zeros(clmns2read,1);
                c = zeros(clmns2read,1);
                for n = 1: clmns2read
                    rtt = rt;
                    ctt = ct + n;
                    r(n) = rtt;
                    c(n) = ctt;
                end
            else
                clmns2read = length(startReading);
                r = zeros(clmns2read,1);
                c = zeros(clmns2read,1);
                for n = 1: clmns2read
                    [rt,ct] = find(strcmp(startReading(n), data_raw));
                    r(n) = rt;
                    c(n) = ct;
                end
            end
            
            
            
            % data integrity check
            t = r(1);
            for n = 1:clmns2read
                if ~isequal(t,r(n))
                    error('Wrong file provided: Data labels not in the same row\n');
                end
            end
            
            tend = size(data_raw,1);
            horizon = tend-r(1);
            clmns = zeros(horizon, clmns2read);
            for n = 1:clmns2read
                for ii = 1:horizon
                    if cellfun(@(x) ischar(x),data_raw(r(n)+ii,c(n)))
                        clmns(ii,n) = cellfun(@(x) str2double(x),data_raw(r(n)+ii,c(n)));
                    elseif cellfun(@(x) isnumeric(x),data_raw(r(n)+ii,c(n)))
                        clmns(ii,n) = cellfun(@(x) (x),data_raw(r(n)+ii,c(n)));
                    end
                end
            end
        end
        
    end
    
end

