% ------------------------------------------------------------------------
% This file is part of the EHCM Toolbox v1
%
% The EHCM Toolbox - Energy Hub Component Modeling for Model Predictive Control.
% Copyright (C) 2015  Automatic Control Laboratory, ETH Zurich.
%
% The EHCM Toolbox is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% The EHCM Toolbox is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with the EHCM Toolbox.  If not, see <http://www.gnu.org/licenses/>.
%
% ------------------------------------------------------------------------
    
function [EH,EHB,COM] = getHubOptProb(EHcomp,EHBcomp,...
    inputPrefix,outputPrefix,Np,distVec,x0EH,x0EHB)


nEHcomp = length(EHcomp);
% find all the different nodes of the EH
nodesList = [];
inputList = [];
for ii = 1:nEHcomp
    teMat = EHcomp{ii}.discreteTimeModel.effMatrix;
    for jj = 1:size(teMat,1)
        tnode = teMat{jj,1};
        [nodesList,inputList] = nodeListUpdate(nodesList,inputList,tnode);
        tnode = teMat{jj,3};
        [nodesList,inputList] = nodeListUpdate(nodesList,inputList,tnode);
    end
end

% find all the different nodes of the EHB
nEHBcomp = length(EHBcomp);
for ii = 1:nEHBcomp
    teMat = EHBcomp{ii}.discreteTimeModel.effMatrix;
    for jj = 1:size(teMat,1)
        tnode = teMat{jj,1};
        [nodesList,inputList] = nodeListUpdate(nodesList,inputList,tnode);
        tnode = teMat{jj,3};
        [nodesList,inputList] = nodeListUpdate(nodesList,inputList,tnode);
    end
end

% assing a control variable to every input over the prediction horizon Np
nu = length(inputList);
exportListVar = sdpvar(nu,Np,'full');
for ii = 1:nu
    if strncmp(inputList{ii},'u_',2)
        eval([inputList{ii},'=',sprintf('exportListVar(%d,:);',ii)]); % outputs explicitly depend on inputs
    end
end

% create a cell array for the nodes of the system
nnod = length(nodesList);
for ii = 1:nnod
    eval([nodesList{ii},'=',sprintf('cell(1,%d);',Np)]);
    for kk = 1:Np
        eval([nodesList{ii},sprintf('{:,%d}',kk),'= 0;']);
    end
end

%% find the input/output/inside(throughput) nodes (verified that works properly)
% input nodes are used for the objective function
% output nodes are used for the interconnection constraint set
% inside(throughput) nodes are used for the EH constraint set

ninput = length(inputPrefix);
inputIdx = [];
for jj = 1:ninput
    inputIdx = [inputIdx;find(strncmp(nodesList,inputPrefix{jj},length(inputPrefix{jj}))==1)];
end

noutput = length(outputPrefix);
outputIdx = [];
for jj = 1:noutput
    outputIdx = [outputIdx;find(strncmp(nodesList,outputPrefix{jj},length(outputPrefix{jj}))==1)];
end

insideIdx = [];
concat = [inputIdx;outputIdx];
for jj = 1:length(nodesList)
    if ~any(concat==jj)
        insideIdx = [insideIdx;jj];
    end
end

%% generate the constraints and objective function of the EH (verified that works properly)

objEH = 0;  % the objective of the energy hub is defined outside, however it can be modified in case that
            % an energy hub component should have some objective function (e.g. heat pump around nominal temperature) 
conEH = [];
xEH = cell(nEHcomp,1);
uEH = cell(nEHcomp,1);
vEH = cell(nEHcomp,1);

for ii = 1:nEHcomp
    
    % find the inputs,states and outputs of the component
    discSS = EHcomp{ii}.discreteTimeModel;
    
    nu = length(discSS.identifiers.u);
    nx = length(discSS.identifiers.x);
    ny = length(discSS.identifiers.y);
    ut = [];
    for jj = 1:nu
        eval(['ut = [ut;',discSS.identifiers.u{jj},'];']);
    end
    
    xt = [x0EH{ii},sdpvar(nx,Np,'full')];
    Ae = discSS.ss.A;
    Be = discSS.ss.B;
    contt = [];
    for jj = 1:Np
        contt = contt + [xt(:,jj+1) == Ae*xt(:,jj) + Be*ut(:,jj)];
    end
    % express outputs as explicit functions of inputs
    yt = [];
    C = discSS.ss.C;
    D = discSS.ss.D;
    for jj = 1:Np
        ytt = C*xt(:,jj) + D*ut(:,jj);
        yt = [yt,ytt];
    end
    % this is needed to get the connection matrix
    for jj = 1:ny
        eval([discSS.identifiers.y{jj},sprintf(' = yt(%d,:);',jj)]);
    end
    
    % constraint set for the specific device
    cont = EHConvConstraints(discSS,xt,ut,Np);
    conEH = conEH + cont + contt;
    
    % return values
    xEH{ii} = xt;
    uEH{ii} = ut;
    
end

% balance and input nodes equations (one column of the efficiency
% matrix is a control variable and the other column is a
% balancing/input/output node. The nodes are cells (e.g. {:,k}) while
% the control variables are sdpvar (e.g. (:,k)). If the control
% varialbe is at the first column then it is added by effCoef times
% to the arrival node at the third column, while if the control variable
% is at the third node then it is subtracted by the departure node with
% 1/eff times.
for ii = 1:nEHcomp
    discSS = EHcomp{ii}.discreteTimeModel;
    effMat = discSS.effMatrix;
    nef = size(effMat,1);
    for jj = 1:nef
        if strncmp(effMat{jj,1},'u_',2) || strncmp(effMat{jj,1},'y_',2)
            for kk = 1:Np
                eval([effMat{jj,3},sprintf('{:,%d}',kk),'=',effMat{jj,3},sprintf('{:,%d}',kk),...
                    sprintf('+ %.4f *',effMat{jj,2}),effMat{jj,1},sprintf('(:,%d)',kk),';']);
            end
        
        elseif strncmp(effMat{jj,3},'u_',2) || strncmp(effMat{jj,3},'y_',2)
            for kk = 1:Np
                eval([effMat{jj,1},sprintf('{:,%d}',kk),'=',effMat{jj,1},sprintf('{:,%d}',kk),...
                    sprintf('- %.4f *',1/effMat{jj,2}),effMat{jj,3},sprintf('(:,%d)',kk),';']);
            end
        else
            error('The efficiency matrix contains not supported entries')
        end
    end
end

% use the inside (balancing) nodes to extend the constraint set
for ii = 1:length(insideIdx)
    for kk = 1:Np
        eval(['conEH = conEH + [',nodesList{insideIdx(ii)},sprintf('{:,%d}',kk),'==0];']);
    end
end


%% generate the constraints and objective function of the EHB

conEHB = cell(nEHBcomp,1);
objEHB = cell(nEHBcomp,1);
xEHB = cell(nEHBcomp,1);
uEHB = cell(nEHBcomp,1);
vEHB = cell(nEHBcomp,1);
for ii = 1:nEHBcomp
    
    discSS = EHBcomp{ii}.discreteTimeModel;
    vbd = distVec{ii};
    
    % building states
    nu = length(discSS.identifiers.u);
    nx = length(discSS.identifiers.x);
    ubd = [];
    for jj = 1:nu
        eval(['ubd = [ubd;',discSS.identifiers.u{jj},'];']);
    end
    xbd = [x0EHB{ii},sdpvar(nx,Np,'full')];
    
    % building constraints (constraints depend on current state and
    % disturbance vector of the building. Therefore, they should be
    % calculated at each iteration.

    conbd = buildingConstraints(EHBcomp{ii},xbd,ubd,vbd,Np);
    
    objbd = 0;
    
    
    conEHB{ii} = conbd;
    objEHB{ii} = objbd;
    xEHB{ii} = xbd;
    uEHB{ii} = ubd;
    vEHB{ii} = vbd.vvar';
end

% efficiency matrix is handled as explained previously
for ii = 1:nEHBcomp
    discSS = EHBcomp{ii}.discreteTimeModel;
    effMat = discSS.effMatrix;
    nef = size(effMat,1);
    for jj = 1:nef
        if strncmp(effMat{jj,1},'u_',2) || strncmp(effMat{jj,1},'y_',2)
            for kk = 1:Np
                eval([effMat{jj,3},sprintf('{:,%d}',kk),'=',effMat{jj,3},sprintf('{:,%d}',kk),...
                    sprintf('+ %.4f *',effMat{jj,2}),effMat{jj,1},sprintf('(:,%d)',kk),';']);
            end
        
        elseif strncmp(effMat{jj,3},'u_',2) || strncmp(effMat{jj,3},'y_',2)
            for kk = 1:Np
                eval([effMat{jj,1},sprintf('{:,%d}',kk),'=',effMat{jj,1},sprintf('{:,%d}',kk),...
                    sprintf('- %.4f *',1/effMat{jj,2}),effMat{jj,3},sprintf('(:,%d)',kk),';']);
            end
        else
            error('The efficiency matrix contains not supported entries')
        end
    end
end

%% generate the interconnection set

% use the output nodes to create the common constraint set
conCOM = [];
for ii = 1:length(outputIdx)
    for kk = 1:Np
        eval(['conCOM = conCOM + [',nodesList{outputIdx(ii)},sprintf('{:,%d}',kk),'==0];']);
    end
end

%% return values

% energy hub
EH.conEH = conEH;
EH.objEH = objEH;
EH.xEH = xEH;
EH.uEH = uEH;
EH.vEH = vEH;


for ii = 1:ninput
    eval(['EH.inputs.',nodesList{inputIdx(ii)},'= [];']);
    % reverse sign to correspond to correct input streams.
    % the minus (-) is used because the input nodes are entry nodes so the
    % control valiables are departing from them and according to convention
    % they should have a minus in front.
    for kk = 1:Np
        eval(['EH.inputs.',nodesList{inputIdx(ii)},'= ',...
            '[ EH.inputs.',nodesList{inputIdx(ii)},...
            ',-1*',nodesList{inputIdx(ii)},sprintf('{:,%d}',kk),'];']);
    end
end

% buildings
EHB.conEHB = conEHB;
EHB.objEHB = objEHB;
EHB.xEHB = xEHB;
EHB.uEHB = uEHB;
EHB.vEHB = vEHB;

% common constraints set
COM.conCOM = conCOM;

end
